import copy
import os
import pickle
import shutil
import subprocess
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from runhouse.globals import rns_client
from runhouse.logger import get_logger

from runhouse.resources.hardware import _current_cluster, _get_cluster_from, Cluster
from runhouse.resources.resource import Resource
from runhouse.rns.utils.api import generate_uuid, relative_file_path
from runhouse.utils import locate_working_dir

logger = get_logger(__name__)


class Folder(Resource):
    RESOURCE_TYPE = "folder"
    DEFAULT_FS = "file"
    CLUSTER_FS = "ssh"
    DEFAULT_FOLDER_PATH = "/runhouse-folder"
    DEFAULT_CACHE_FOLDER = "~/.cache/runhouse"

    def __init__(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        system: Union[str, Cluster] = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Folder object.

        .. note::
            To build a folder, please use the factory method :func:`folder`.
        """
        super().__init__(name=name, dryrun=dryrun)

        # https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.gui.FileSelector.urlpath
        # Note: no longer needed as part of previous fsspec usage, but still used by some s3 / gsutil commands
        self._urlpath = None

        self._system = _get_cluster_from(
            system or _current_cluster(key="config"), dryrun=dryrun
        )

        # TODO [DG] Should we ever be allowing this to be None?
        if path is None:
            self._path = Folder.default_path(self.rns_address, self._system)
        else:
            if self._system != self.DEFAULT_FS:
                self._path = path
            else:
                self._path = self._path_absolute_to_rh_workdir(path)

    def __getstate__(self):
        """Override the pickle method to clear _urlpath before pickling."""
        state = self.__dict__.copy()
        state["_urlpath"] = None
        return state

    @classmethod
    def default_path(cls, rns_address, system):
        name = (
            rns_client.split_rns_name_and_path(rns_address)[0]
            if rns_address
            else generate_uuid()
        )

        if system == Folder.DEFAULT_FS:
            return str(Path.cwd() / name)
        elif isinstance(system, Cluster):
            return f"{Folder.DEFAULT_CACHE_FOLDER}/{name}"
        else:
            return f"{Folder.DEFAULT_FOLDER_PATH}/{name}"

    # ----------------------------------
    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        if _resolve_children:
            config = Folder._check_for_child_configs(config)

        if config["system"] == "s3":
            from .s3_folder import S3Folder

            return S3Folder.from_config(config, dryrun=dryrun)
        elif config["system"] == "gs":
            from .gcs_folder import GCSFolder

            return GCSFolder.from_config(config, dryrun=dryrun)
        elif isinstance(config["system"], dict):
            config["system"] = Cluster.from_config(
                config["system"], dryrun=dryrun, _resolve_children=_resolve_children
            )
        return Folder(**config, dryrun=dryrun)

    @classmethod
    def _check_for_child_configs(cls, config: Dict):
        """Overload by child resources to load any resources they hold internally."""
        system = config.get("system")
        if isinstance(system, str) or isinstance(system, dict):
            config["system"] = _get_cluster_from(system)
        return config

    @property
    def path(self):
        """Folder path."""
        if self._path is not None:
            if self.system == Folder.DEFAULT_FS:
                return str(Path(self._path).expanduser())
            return str(self._path)
        else:
            return None

    @path.setter
    def path(self, path):
        self._path = path

    @property
    def system(self):
        """Filesystem or cluster that the folder is on."""
        return self._system

    @system.setter
    def system(self, new_system: Union[str, Cluster]):
        self._system = _get_cluster_from(new_system)

    @property
    def _fs_str(self):
        if isinstance(self.system, Cluster):
            if self.system.on_this_cluster():
                return self.DEFAULT_FS
            return self.CLUSTER_FS
        else:
            return self.system

    @property
    def local_path(self):
        """Local path of the folder, if it exists locally. Else, ``None``."""
        if self.is_local():
            return str(Path(self.path).expanduser())
        else:
            return None

    @property
    def _use_http_endpoint(self):
        """Whether to use system APIs to perform folder operations on the cluster via HTTP."""
        return isinstance(self.system, Cluster) and not self.system.on_this_cluster()

    def mv(
        self,
        system: Union[str, Cluster],
        path: Optional[str],
        overwrite: bool = True,
    ) -> None:
        """Move the folder to a new filesystem or cluster.

        Args:
            system (str or Cluster): Filesystem or cluster to move the folder to.
            path (str): Path to move the folder to.
            overwrite (bool, optional): Whether to override if a file already exists at the destination path.
                (Default: ``True``)

        Example:
            >>> folder = rh.folder(path="local/path")
            >>> folder.mv(my_cluster)
            >>> folder.mv("s3", "s3_bucket/path")
        """
        if path is None:
            raise ValueError("A destination path must be specified.")

        src_path = Path(self.path)
        dest_path = Path(path)

        if self._use_http_endpoint:
            self.system._folder_mv(
                path=src_path, dest_path=dest_path, overwrite=overwrite
            )
        else:
            dest_path = dest_path.expanduser()
            src_path = src_path.expanduser()

            if not src_path.exists():
                raise FileNotFoundError(f"The source path {src_path} does not exist.")

            if system == self.DEFAULT_FS:
                if dest_path.exists():
                    raise FileExistsError(
                        f"The destination path {dest_path} already exists."
                    )

                # Create the destination directory if it doesn't exist
                dest_path.parent.mkdir(parents=True, exist_ok=overwrite)

                # Move the directory
                shutil.move(str(src_path), str(dest_path))

                # Update the path attribute
                self.path = str(dest_path)
                self.system = self.DEFAULT_FS

            else:
                # TODO [JL] support moving to other systems
                raise NotImplementedError(f"System {system} not supported for local mv")

    def to(
        self,
        system: Union[str, Cluster],
        path: Optional[Union[str, Path]] = None,
    ):
        """Copy the folder to a new filesystem or cluster.
        Currently supported: ``here``, ``file``, ``gs``, ``s3``, or a cluster.

        Args:
            system (str or Cluster): Filesystem or cluster to move the folder to.
            path (str, optional): Path to move the folder to.

        Example:
            >>> local_folder = rh.folder(path="/my/local/folder")
            >>> s3_folder = local_folder.to("s3")
        """
        if system == "here":
            current_cluster_config = _current_cluster(key="config")
            if current_cluster_config:
                system = Cluster.from_config(current_cluster_config)
            else:
                system = self.DEFAULT_FS
            path = str(Path.cwd() / self.path.split("/")[-1]) if path is None else path

        if isinstance(system, Cluster):
            if not path:
                # Use a default path on the cluster (in ~/.cache/runhouse)
                dest_path = Folder.default_path(
                    rns_address=self.rns_address, system=system
                )
            else:
                # Destination path on the cluster should be a relative path
                dest_path = (
                    relative_file_path(file_path=path)
                    if Path(path).is_absolute()
                    else path
                )

            # rsync the folder contents to the cluster's destination path
            logger.debug(f"Syncing folder contents to cluster in path: {dest_path}")
            return self._to_cluster(system, path=dest_path)

        path = str(
            path or Folder.default_path(self.rns_address, system)
        )  # Make sure it's a string and not a Path

        system_str = getattr(
            system, "name", system
        )  # Use system.name if available, i.e. system is a cluster
        logger.info(
            f"Copying folder from {self.fsspec_url} to: {system_str}, with path: {path}"
        )

        # to_local, to_cluster and to_data_store are also overridden by subclasses to dispatch
        # to more performant cloud-specific APIs
        system = _get_cluster_from(system)
        if system == "file":
            return self._to_local(dest_path=path)
        elif system in ["s3", "gs"]:
            return self._to_data_store(system=system, data_store_path=path)
        else:
            raise ValueError(
                f"System '{system}' not currently supported as a destination system."
            )

    def _destination_folder(
        self,
        dest_path: str,
        dest_system: Optional[str] = "file",
    ):
        """Returns a new Folder object pointing to the destination folder."""
        folder_config = self.config()
        folder_config["system"] = dest_system
        folder_config["path"] = dest_path
        new_folder = Folder.from_config(folder_config)

        return new_folder

    def _to_local(self, dest_path: str):
        """Copies folder to local. Only relevant for the base Folder if its system is a cluster."""
        if isinstance(self.system, Cluster):
            # Cluster --> local copying
            logger.debug(
                f"Copying folder from cluster {self.system.name} to local path: {dest_path}"
            )
            # Return a new folder objecting pointing to the local destination path
            return self._cluster_to_local(self.system, dest_path)

        if self.system == self.DEFAULT_FS:
            # Local --> local copying
            logger.debug(f"Copying folder to local path: {dest_path}")
            self.mv(system=self.system, path=dest_path)
            return self

        raise TypeError(f"Cannot copy from {self.system} to local.")

    def _to_data_store(
        self,
        system: str,
        data_store_path: Optional[str] = None,
    ):
        """Local or cluster to blob storage."""
        local_folder_path = self.path

        # The new folder should be a sub-folder for the relevant data store (e.g. `S3Folder`)
        new_folder = self._destination_folder(
            dest_path=data_store_path, dest_system=system
        )
        new_folder._upload(src=local_folder_path)
        return new_folder

    def mkdir(self):
        """Create the folder in specified file system if it doesn't already exist."""
        if self._use_http_endpoint:
            self.system._folder_mkdir(self.path)
        else:
            path = Path(self.path).expanduser()
            if not path.exists():
                path.mkdir(parents=True, exist_ok=True)

        logger.debug(f"Folder created in path: {self.path}")

    def _to_cluster(self, dest_cluster, path=None):
        """Copy the folder from a file or cluster source onto a destination cluster."""
        if not dest_cluster.ips:
            raise ValueError("Cluster must be started before copying data to it.")

        dest_path = path or f"~/{Path(self.path).name}"

        # Need to add slash for rsync to copy the contents of the folder
        dest_folder = copy.deepcopy(self)
        dest_folder.path = dest_path
        dest_folder.system = dest_cluster

        if self._fs_str == self.DEFAULT_FS and dest_cluster.name is not None:
            # Includes case where we're on the cluster itself
            # And the destination is a cluster, not rh.here
            dest_cluster.rsync(source=self.path, dest=dest_path, up=True, contents=True)

        elif isinstance(self.system, Resource):
            if self.system.endpoint(external=False) == dest_cluster.endpoint(
                external=False
            ):
                # We're on the same cluster, so we can just move the files
                if not path:
                    # If user didn't specify a path, we can just return self
                    return self
                else:
                    dest_cluster.run_bash(
                        [f"mkdir -p {dest_path}", f"cp -r {self.path}/* {dest_path}"],
                    )
            else:
                self._cluster_to_cluster(dest_cluster, dest_path)

        else:
            # data store folders have their own specific _to_cluster functions
            raise TypeError(
                f"`Sending from filesystem type {type(self.system)} is not supported"
            )

        return dest_folder

    def _cluster_to_cluster(self, dest_cluster, dest_path):
        src_path = self.path

        cluster_creds = self.system.creds_values

        if not cluster_creds.get("password") and not dest_cluster.creds_values.get(
            "password"
        ):
            creds_file = cluster_creds.get("ssh_private_key")
            creds_cmd = f"-i '{creds_file}' " if creds_file else ""

            dest_cluster.run_bash([f"mkdir -p {dest_path}"])
            command = (
                f"rsync -Pavz --filter='dir-merge,- .gitignore' -e \"ssh {creds_cmd}"
                f"-o StrictHostKeyChecking=no -o IdentitiesOnly=yes -o ExitOnForwardFailure=yes "
                f"-o ServerAliveInterval=5 -o ServerAliveCountMax=3 -o ConnectTimeout=30s -o ForwardAgent=yes "
                f'-o ControlMaster=auto -o ControlPersist=300s" {src_path}/ {dest_cluster.head_ip}:{dest_path}'
            )
            status_codes = self.system.run_bash([command])
            if status_codes[0][0] != 0:
                raise Exception(
                    f"Error syncing folder to destination cluster ({dest_cluster.name}). "
                    f"Make sure the source cluster ({self.system.name}) has the necessary provider keys "
                    f"if applicable. "
                )
        else:
            local_folder = self._cluster_to_local(self.system, self.path)
            local_folder._to_cluster(dest_cluster, dest_path)

    def _cluster_to_local(self, cluster, dest_path):
        """Create a local folder with dest_path from the cluster.

        This function rsyncs down the data and return a folder with system=='file'.
        """
        if not cluster.ips:
            raise ValueError("Cluster must be started before copying data from it.")
        Path(dest_path).expanduser().mkdir(parents=True, exist_ok=True)
        cluster.rsync(
            source=self.path,
            dest=str(Path(dest_path).expanduser()),
            up=False,
            contents=True,
        )
        new_folder = copy.deepcopy(self)
        new_folder.path = dest_path
        new_folder.system = self.DEFAULT_FS
        return new_folder

    def is_local(self):
        """Whether the folder is on the local filesystem.

        Example:
            >>> is_local = my_folder.is_local()
        """
        return (
            self._fs_str == self.DEFAULT_FS
            and self.path is not None
            and Path(self.path).expanduser().exists()
        )

    def _upload(self, src: str, region: Optional[str] = None):
        """Upload a folder to a remote folder."""
        raise NotImplementedError

    def _upload_command(self, src: str, dest: str):
        """CLI command for uploading folder to remote bucket. Needed when uploading a folder from a cluster."""
        raise NotImplementedError

    def _upload_folder_to_bucket(self, command: str):
        """Uploads a folder to a remote bucket (e.g. s3).
        Based on the CLI command skypilot uses to upload the folder"""
        # Adapted from: https://github.com/skypilot-org/skypilot/blob/983f5fa3197fe7c4b5a28be240f7b027f7192b15/sky/data/data_utils.py#L165 # noqa
        with subprocess.Popen(
            command, stderr=subprocess.PIPE, stdout=subprocess.DEVNULL, shell=True
        ) as process:
            stderr = []
            while True:
                line = process.stderr.readline()
                if not line:
                    break
                str_line = line.decode("utf-8")
                stderr.append(str_line)
            returncode = process.wait()
            if returncode != 0:
                raise RuntimeError(" ".join(stderr).strip())

    def _download(self, dest):
        raise NotImplementedError

    def _download_command(self, src, dest):
        """CLI command for downloading folder from remote bucket. Needed when downloading a folder to a cluster."""
        raise NotImplementedError

    def config(self, condensed: bool = True):
        config = super().config(condensed)

        if self.system == Folder.DEFAULT_FS:
            # If folder is local check whether path is relative, and if so take it relative to the working director
            # rather than to the home directory. If absolute, it's left alone.
            config["path"] = (
                self._path_relative_to_rh_workdir(self.path) if self.path else None
            )
        else:
            # if not a local filesystem save path as is (e.g. bucket/path)
            config["path"] = self.path

        if isinstance(self.system, Resource):  # If system is a cluster
            config["system"] = self._resource_string_for_subconfig(
                self.system, condensed
            )
        else:
            config["system"] = self.system

        return config

    def _save_sub_resources(self, folder: str = None):
        if isinstance(self.system, Resource):
            self.system.save(folder=folder)

    @staticmethod
    def _path_relative_to_rh_workdir(path: str):
        rh_workdir = Path(locate_working_dir())
        try:
            return str(Path(path).relative_to(rh_workdir))
        except ValueError:
            return path

    @staticmethod
    def _path_absolute_to_rh_workdir(path: str):
        return (
            path
            if Path(path).expanduser().is_absolute()
            else str(Path(locate_working_dir()) / path)
        )

    @staticmethod
    def _delete_contents(contents: List, folder_path: Path, recursive: bool):
        for content in contents:
            content_path = folder_path / content
            if content_path.exists():
                if content_path.is_file():
                    content_path.unlink()
                elif content_path.is_dir() and recursive:
                    shutil.rmtree(content_path)
                else:
                    raise ValueError(
                        f"Path {content_path} is a directory and recursive is set to False"
                    )

    @property
    def fsspec_url(self):
        """Generate the FSSpec style URL using the file system and path of the folder"""
        if self.path.startswith("/") and self._fs_str not in [
            rns_client.DEFAULT_FS,
            self.CLUSTER_FS,
        ]:
            return f"{self._fs_str}:/{self.path}"
        else:
            # For local, ssh / sftp filesystems we need both slashes
            # e.g.: 'ssh:///home/ubuntu/.cache/runhouse/tables/dede71ef83ce45ffa8cb27d746f97ee8'
            return f"{self._fs_str}://{self.path}"

    @property
    def _bucket_name(self):
        return self.path.lstrip("/").split("/")[0]

    @property
    def _key(self):
        filtered_parts = self.path.split("/")[2:]
        return "/".join(filtered_parts) + "/"

    def ls(self, full_paths: bool = True, sort: bool = False) -> List:
        """List the contents of the folder.

        Args:
            full_paths (bool, optional): Whether to list the full paths of the folder contents.
                Defaults to ``True``.
            sort (bool, optional): Whether to sort the folder contents by time modified.
                Defaults to ``False``.
        """
        if self._use_http_endpoint:
            return self.system._folder_ls(self.path, full_paths=full_paths, sort=sort)
        else:
            path = Path(self.path).expanduser()
            paths = [p for p in path.iterdir()]

            # Sort the paths by modification time if sort is True
            if sort:
                paths.sort(key=lambda p: p.stat().st_mtime, reverse=True)

            # Convert paths to strings and format them based on full_paths
            if full_paths:
                return [str(p.resolve()) for p in paths]
            else:
                return [p.name for p in paths]

    def resources(self, full_paths: bool = False):
        """List the resources in the folder.

        Args:
            full_paths (bool, optional): Whether to list the full path or relative path. (Default: ``False``)

        Example:
            >>> resources = my_folder.resources()
        """
        try:

            if self._use_http_endpoint:
                paths = self.system._folder_ls(path=self.path, full_paths=full_paths)
                resources = [
                    path
                    for path in paths
                    if self.system._folder_exists(path=f"{path}/config.json")
                ]
                return resources
            else:
                resources = [
                    path for path in self.ls() if (Path(path) / "config.json").exists()
                ]
                if full_paths:
                    return [
                        self.rns_address + "/" + Path(path).stem for path in resources
                    ]
                else:
                    return [Path(path).stem for path in resources]

        except Exception as e:
            logger.error(f"Error listing resources in folder: {e}")
            return []

    @property
    def rns_address(self):
        # TODO Maybe later, account for folders along the path with a different RNS name.

        if self.name is None:  # Anonymous folders have no rns address
            return None

        # Only should be necessary when a new base folder is being added (therefore isn't in rns_base_folders yet)
        if self._rns_folder:
            return str(Path(self._rns_folder) / self.name)

        if self.path in rns_client.rns_base_folders.values():
            if self._rns_folder:
                return self._rns_folder + "/" + self.name
            else:
                return rns_client.default_folder + "/" + self.name

        segment = Path(self.path)
        while (
            str(segment) not in rns_client.rns_base_folders.values()
            and not segment == Path.home()
            and not segment == segment.parent
        ):
            segment = segment.parent

        if (
            segment == Path.home() or segment == segment.parent
        ):  # TODO throw an error instead?
            return rns_client.default_folder + "/" + self.name
        else:
            base_folder = Folder(path=str(segment), dryrun=True)
            base_folder_path = base_folder.rns_address
            relative_path = str(Path(self.path).relative_to(base_folder.path))
            return base_folder_path + "/" + relative_path

    def contains(self, name_or_path: str) -> bool:
        """Whether path exists locally inside a folder.

        Args:
            name_or_path (str): Name or path of folder to check if it exists inside the folder.

        Example:
            >>> my_folder = rh.folder("local/folder/path")
            >>> in_folder = my_folder.contains("filename")
        """
        path, _ = self.locate(name_or_path)
        return path is not None

    def locate(self, name_or_path) -> Tuple[str, str]:
        """Locate the local path of a Folder given an rns path.

        Args:
            name_or_path (str): Name or path of folder to locate inside the folder.

        Example:
            >>> my_folder = rh.folder("local/folder/path")
            >>> local_path = my_folder.locate("file_name")
        """
        # Note: Keep in mind we're using both _rns_ path and physical path logic below. Be careful!

        # If the path is already given relative to the current folder:
        if (Path(self.path) / name_or_path).exists():
            return str(Path(self.path) / name_or_path), self.system

        # If name or path uses ~/ or ./, need to resolve with folder path
        abs_path = rns_client.resolve_rns_path(name_or_path)
        rns_path = self.rns_address

        # If this folder is anonymous, it has no rns contents
        if rns_path is None:
            return None, None

        if abs_path == rns_path:
            return self.path, self.system
        try:
            child_path = Path(self.path) / Path(abs_path).relative_to(rns_path)
            if child_path.exists():
                return str(child_path), self.system
        except ValueError:
            pass

        # Last resort, recursively search inside sub-folders and children.

        segments = abs_path.lstrip("/").split("/")
        if len(segments) == 1:
            return (
                None,
                None,
            )  # If only a single element, would have been found in ls above.

        # Look for lowest folder in the path that exists in filesystem, and recurse from that folder
        greatest_common_folder = Path(self.path)
        i = 0
        for i, seg in enumerate(segments):
            if not (greatest_common_folder / seg).exists():
                break
            greatest_common_folder = greatest_common_folder / seg
        if not str(greatest_common_folder) == self.path:
            return Folder(
                path=str(greatest_common_folder), system=self.system, dryrun=True
            ).locate("/".join(segments[i + 1 :]))

        return None, None

    def open(self, name, mode: str = "rb", encoding: Optional[str] = None):
        """Returns the specified file as a stream (`botocore.response.StreamingBody`), which must be used as a
        content manager to be opened.

        Args:
            name (str): Name of file inside the folder to open.
            mode (str, optional): Mode for opening the file. (Default: ``"rb"``)
            encoding (str, optional): Encoding for opening the file. (Default: ``None``)

        Example:
            >>> with my_folder.open('obj_name') as my_file:
            >>>        pickle.load(my_file)
        """
        file_path = Path(self.path).expanduser() / name
        valid_modes = {"r", "w", "a", "rb", "wb", "ab", "r+", "w+", "a+"}

        if mode not in valid_modes:
            raise NotImplementedError(
                f"{mode} mode is not implemented yet for local files"
            )

        return open(file_path, mode=mode, encoding=encoding)

    def get(self, name, mode: str = "rb", encoding: Optional[str] = None):
        """Returns the contents of a file as a string or bytes.

        Args:
            name (str): Name of file to get contents of.
            mode (str, optional): Mode for opening the file. (Default: ``"rb"``)
            encoding (str, optional): Encoding for opening the file. (Default: ``None``)

        Example:
            >>> contents = my_folder.get(file_name)
        """
        if self._use_http_endpoint:
            return self.system._folder_get(
                path=Path(self.path) / name, mode=mode, encoding=encoding
            )

        with self.open(name, mode=mode, encoding=encoding) as f:
            return f.read()

    # TODO [DG] fix this to follow the correct convention above
    def get_all(self):
        # TODO add docs for this
        # TODO we're not closing these, do we need to extract file-like objects so we can close them?
        raise NotImplementedError

    def exists_in_system(self):
        """Whether the folder exists in the filesystem.

        Example:
            >>> exists_on_system = my_folder.exists_in_system()
        """
        if self._use_http_endpoint:
            return self.system._folder_exists(path=self.path)
        else:
            full_path = Path(self.path).expanduser()
            return full_path.exists() and full_path.is_dir()

    def rm(self, contents: List = None, recursive: bool = True):
        """Delete a folder from the file system. Optionally provide a list of folder contents to delete.

        Args:
            contents (Optional[List]): Specific contents to delete in the folder. If None, removes
                the entire folder. (Default: ``None``)
            recursive (bool, optional): Delete the folder itself (including all its contents).
                (Default: ``True``)

        Example:
            >>> my_folder.rm()
        """
        if self._use_http_endpoint:
            self.system._folder_rm(self.path, contents, recursive)
        else:
            folder_path = Path(self.path).expanduser()
            if contents:
                Folder._delete_contents(contents, folder_path, recursive)
            else:
                if recursive:
                    shutil.rmtree(folder_path)
                else:
                    if folder_path.is_dir():
                        for item in folder_path.iterdir():
                            if item.is_file():
                                item.unlink()
                            else:
                                raise ValueError(
                                    f"Folder {item} found in {folder_path}, recursive is set to False"
                                )
                    else:
                        folder_path.unlink()

    def put(self, contents: Dict[str, Any], overwrite: bool = False, mode: str = "wb"):
        """Put given contents in folder.

        Args:
            contents (Dict[str, Any] or Resource or List[Resource]): Contents to put in folder.
                Must be a dict with keys being the file names (without full paths) and values being the file-like
                objects to write, or a Resource object, or a list of Resources.
            overwrite (bool, optional): Whether to overwrite the existing file if it exists. (Default: ``False``)
            mode (str, optional): Write mode to use. (Default: ``wb``).

        Example:
            >>> my_folder.put(contents={"filename.txt": data})
        """
        # Handle lists of resources just for convenience
        if isinstance(contents, list):
            for resource in contents:
                self.put(resource, overwrite=overwrite)
            return

        if self._use_http_endpoint:
            self.system._folder_put(
                path=self.path, contents=contents, overwrite=overwrite, mode=mode
            )
        else:
            self.mkdir()
            full_path = str(Path(self.path).expanduser())
            if isinstance(contents, Folder):
                if (
                    contents.path is None
                ):  # Should only be the case when Folder is created
                    contents.path = os.path.join(full_path, contents.name)
                return

            if not isinstance(contents, dict):
                raise TypeError(
                    "`contents` argument must be a dict mapping filenames to file-like objects"
                )

            if overwrite is False:
                # Check if files exist and raise an error if they do
                existing_files = set(os.listdir(full_path))
                intersection = existing_files.intersection(set(contents.keys()))
                if intersection:
                    raise FileExistsError(
                        f"File(s) {intersection} already exist(s) at path: {full_path}. "
                        f"Cannot save them with overwrite={overwrite}."
                    )

            for filename, file_obj in contents.items():
                file_obj = self._serialize_file_obj(file_obj)
                file_path = Path(full_path) / filename
                if not overwrite and file_path.exists():
                    raise FileExistsError(f"File {file_path} already exists.")

                try:
                    with open(file_path, mode) as f:
                        f.write(file_obj)

                except Exception as e:
                    raise RuntimeError(
                        f"Failed to write {filename} to {file_path}: {e}"
                    )

    @staticmethod
    def _serialize_file_obj(file_obj):
        if not isinstance(file_obj, bytes):
            try:
                file_obj = pickle.dumps(file_obj)
            except (pickle.PicklingError, TypeError) as e:
                raise ValueError(f"Cannot serialize file contents: {e}")

        return file_obj

    @staticmethod
    def _bucket_name_from_path(path: str) -> str:
        """Extract the bucket name from a path (e.g. '/my-bucket/my-folder/my-file.txt' -> 'my-bucket')"""
        return Path(path).parts[1]



import copy
from pathlib import Path

import runhouse as rh
from runhouse.resources.hardware import OnDemandCluster
from runhouse.resources.hardware.utils import ServerConnectionType
from runhouse.rns.utils.api import resolve_absolute_path


def is_on_cluster(cluster):
    return cluster.on_this_cluster()


def np_array(num_list: list):
    import numpy as np

    return np.array(num_list)


def sd_generate_image(prompt):
    from diffusers import StableDiffusionPipeline

    model = StableDiffusionPipeline.from_pretrained(
        "stabilityai/stable-diffusion-2-base"
    ).to("cuda")
    return model(prompt).images[0]


def test_cluster_config(local_launched_ondemand_aws_docker_cluster):
    config = local_launched_ondemand_aws_docker_cluster.config()
    cluster2 = OnDemandCluster.from_config(config)
    assert cluster2.head_ip == local_launched_ondemand_aws_docker_cluster.head_ip


def test_cluster_sharing(local_launched_ondemand_aws_docker_cluster):
    local_launched_ondemand_aws_docker_cluster.share(
        users=["donny@run.house", "josh@run.house"],
        access_level="write",
        notify_users=False,
    )
    assert True


def test_read_shared_cluster(local_launched_ondemand_aws_docker_cluster):
    res = local_launched_ondemand_aws_docker_cluster.run_python(
        ["import numpy", "print(numpy.__version__)"]
    )
    assert res[0][1]


def test_install(cluster):
    cluster.install_packages(
        [
            "./",
            "torch==1.12.1",
            # 'conda:jupyterlab',
            # 'gh:pytorch/vision'
        ]
    )


def test_basic_run(cluster):
    # Create temp file where fn's will be stored
    test_cmd = "echo hi"
    res = cluster.run_bash(commands=[test_cmd])
    assert "hi" in res[0][1]


def test_restart_server(cluster):
    cluster.up_if_not()
    codes = cluster.restart_server(resync_rh=False)
    assert codes


def test_on_same_cluster(cluster):
    hw_copy = copy.copy(cluster)

    func_hw = rh.function(is_on_cluster).to(cluster)
    assert func_hw(cluster)
    assert func_hw(hw_copy)


def test_on_diff_cluster(
    local_launched_ondemand_aws_docker_cluster, static_cpu_pwd_cluster
):
    func_hw = rh.function(is_on_cluster).to(local_launched_ondemand_aws_docker_cluster)
    assert not func_hw(static_cpu_pwd_cluster)


def test_byo_cluster(static_cpu_pwd_cluster, local_folder):
    from tests.test_resources.test_modules.test_functions.conftest import summer

    assert static_cpu_pwd_cluster.is_up()

    summer_func = rh.function(summer).to(static_cpu_pwd_cluster)
    assert summer_func(1, 2) == 3

    static_cpu_pwd_cluster.put("test_obj", list(range(10)))
    assert static_cpu_pwd_cluster.get("test_obj") == list(range(10))

    local_folder = local_folder.to(static_cpu_pwd_cluster)
    assert "sample_file_0.txt" in local_folder.ls(full_paths=False)


def test_byo_cluster_with_https(static_cpu_pwd_cluster):
    tls_connection = ServerConnectionType.TLS.value
    static_cpu_pwd_cluster.server_connection_type = tls_connection
    static_cpu_pwd_cluster.restart_server()

    assert static_cpu_pwd_cluster.server_connection_type == tls_connection

    local_cert_path = static_cpu_pwd_cluster.cert_config.cert_path
    assert Path(local_cert_path).exists()

    # Confirm we can send https requests to the cluster
    static_cpu_pwd_cluster.install_packages(["numpy"])


def test_byo_proxy(static_cpu_pwd_cluster, local_folder):
    from tests.test_resources.test_modules.test_functions.conftest import summer

    rh.globals.open_cluster_tunnels.pop(static_cpu_pwd_cluster.head_ip)
    static_cpu_pwd_cluster.client = None
    # static_cpu_pwd_cluster._rpc_tunnel.close()
    static_cpu_pwd_cluster._rpc_tunnel = None

    static_cpu_pwd_cluster._creds["ssh_host"] = "127.0.0.1"
    static_cpu_pwd_cluster._creds.update(
        {"ssh_proxy_command": "ssh -W %h:%p ubuntu@test-byo-cluster"}
    )
    assert static_cpu_pwd_cluster.up_if_not()

    status, stdout, _ = static_cpu_pwd_cluster.run_bash(["echo hi"])[0]
    assert status == 0
    assert stdout == "hi\n"

    summer_func = rh.function(summer).to(static_cpu_pwd_cluster)
    assert summer_func(1, 2) == 3

    static_cpu_pwd_cluster.put("test_obj", list(range(10)))
    assert static_cpu_pwd_cluster.get("test_obj") == list(range(10))

    # TODO: uncomment out when in-mem lands
    # local_folder = local_folder.to(static_cpu_pwd_cluster)
    # assert "sample_file_0.txt" in local_folder.ls(full_paths=False)


def test_cluster_with_den_auth(
    ondemand_aws_https_cluster_with_auth, summer_func_with_auth
):
    ondemand_aws_https_cluster_with_auth.restart_server()
    from runhouse.globals import configs

    # Create an invalid token, confirm the server does not accept the request
    orig_token = configs.token

    # Request should return 200 using a valid token
    summer_func_with_auth(1, 2)

    configs.token = "abcd123"

    try:
        # Request should raise an exception with an invalid token
        summer_func_with_auth(1, 2)
    except ValueError as e:
        assert "Invalid or expired token" in str(e)

    configs.token = orig_token


def test_start_server_with_custom_certs(
    ondemand_aws_https_cluster_with_auth, summer_func_with_auth
):
    # NOTE: to check certificate matching:
    # openssl x509 -noout -modulus -in rh_server.crt | openssl md5
    # openssl rsa -noout -modulus -in rh_server.key | openssl md5
    from runhouse.servers.http.certs import TLSCertConfig

    ssl_certfile = (
        f"~/ssl/certs/{ondemand_aws_https_cluster_with_auth.name}/rh_server.crt"
    )
    ssl_keyfile = (
        f"~/ssl/private/{ondemand_aws_https_cluster_with_auth.name}/rh_server.key"
    )

    # # NOTE: need to include the IP of the cluster when generating the cert
    TLSCertConfig(
        key_path=ssl_keyfile,
        cert_path=ssl_certfile,
    ).generate_certs(address=ondemand_aws_https_cluster_with_auth.head_ip)

    # # Restart the server using the custom certs
    ondemand_aws_https_cluster_with_auth.ssl_certfile = ssl_certfile
    ondemand_aws_https_cluster_with_auth.ssl_keyfile = ssl_keyfile
    ondemand_aws_https_cluster_with_auth.restart_server()

    try:
        summer_func_with_auth(1, 2)
    except Exception as e:
        assert False, f"Failed to connect to server with custom certs: {e}"

    Path(resolve_absolute_path(ssl_certfile)).unlink()
    Path(resolve_absolute_path(ssl_keyfile)).unlink()



from runhouse.resources.secrets.provider_secrets.api_key_secret import ApiKeySecret


class OpenAISecret(ApiKeySecret):
    """
    .. note::
            To create an OpenAISecret, please use the factory method :func:`provider_secret` with ``provider="openai"``.
    """

    _PROVIDER = "openai"
    _DEFAULT_ENV_VARS = {"api_key": "OPENAI_API_KEY"}

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return OpenAISecret(**config, dryrun=dryrun)



import json

import pytest
import runhouse as rh

from tests.conftest import init_args

from tests.constants import TEST_ORG
from tests.utils import friend_account, friend_account_in_org


def load_shared_resource_config(resource_class_name, address):
    resource_class = getattr(rh, resource_class_name)
    loaded_resource = resource_class.from_name(address, dryrun=True)
    config = loaded_resource.config()
    return config
    # TODO allow resource subclass tests to extend set of properties to test


class TestResource:

    UNIT = {
        "resource": [
            "unnamed_resource",
            "named_resource_for_org",
            "named_resource",
            "local_named_resource",
        ]
    }
    LOCAL = {
        "resource": [
            "unnamed_resource",
            "named_resource_for_org",
            "named_resource",
            "local_named_resource",
        ]
    }
    MINIMAL = {"resource": ["named_resource_for_org", "named_resource"]}
    RELEASE = {
        "resource": [
            "unnamed_resource",
            "named_resource_for_org",
            "named_resource",
            "local_named_resource",
        ]
    }
    MAXIMAL = {
        "resource": [
            "unnamed_resource",
            "named_resource_for_org",
            "named_resource",
            "local_named_resource",
        ]
    }

    @pytest.mark.level("unit")
    def test_resource_factory_and_properties(self, resource):
        assert isinstance(resource, rh.Resource)
        args = init_args.get(id(resource))
        if "name" in args:

            if args["name"].startswith("^"):
                args["name"] = args["name"][1:]

            if "/" in args["name"]:
                assert resource.rns_address == args["name"]
            else:
                assert resource.name == args["name"]
                if resource.rns_address:
                    assert resource.rns_address.split("/")[-1] == args["name"]

        if "dryrun" in args:
            assert args["dryrun"] == resource.dryrun

        assert resource.RESOURCE_TYPE is not None

    @pytest.mark.level("unit")
    def test_config(self, resource):
        args = init_args.get(id(resource))
        config = resource.config()
        assert isinstance(config, dict)
        if resource.rns_address:
            assert config["name"] == resource.rns_address
        else:
            assert config["name"] == resource.name
        assert config["resource_type"] == resource.RESOURCE_TYPE
        assert config["resource_subtype"] == resource.__class__.__name__
        if "dryrun" in args:
            assert config["dryrun"] == args["dryrun"]

    @pytest.mark.level("unit")
    def test_from_config(self, resource):
        new_resource = rh.Resource.from_config(resource.config())
        assert new_resource.config() == resource.config()
        assert new_resource.rns_address == resource.rns_address
        # TODO allow resource subclass tests to extend set of properties to test

    @pytest.mark.level("local")
    def test_save_and_load(self, saved_resource):
        if TEST_ORG in saved_resource.rns_address:
            # Kitchen tester not in the org, so skip for org resources
            return

        alt_resource = None
        # Test loading from name
        loaded_resource = saved_resource.__class__.from_name(saved_resource.rns_address)

        assert loaded_resource.config() == saved_resource.config()
        if isinstance(saved_resource, rh.OnDemandCluster):
            return

        # Do everything inside a try/finally so we don't leave resources behind if the test fails
        try:
            # Test saving under new name
            original_name = saved_resource.rns_address
            alt_name = saved_resource.rns_address + "-alt"

            # This saves a new Den config with the same resource,
            # but an alt name. It also updates the local config to point to the new Den config.
            saved_resource.save(alt_name)
            alt_resource = saved_resource.__class__.from_name(alt_name)
            assert alt_resource.config() == saved_resource.config()

            # Test that original resource is still available in Den
            reloaded_resource = saved_resource.__class__.from_name(original_name)
            assert reloaded_resource.rns_address != alt_resource.rns_address

            # Rename saved resource locally back to original name
            saved_resource.name = original_name
            assert (
                reloaded_resource.rns_address
                == saved_resource.rns_address
                == original_name
            )

        finally:
            if alt_resource is not None:
                alt_resource.delete_configs()
                assert not rh.exists(alt_resource.rns_address)

    @pytest.mark.level("unit")
    def test_history(self, saved_resource):
        if saved_resource.rns_address[:2] == "~/":
            with pytest.raises(ValueError):
                saved_resource.history()
            return

        history = saved_resource.history()
        assert isinstance(history, list)
        assert isinstance(history[0], dict)
        assert "timestamp" in history[0]
        assert "owner" in history[0]
        assert "data" in history[0]
        # Not all config values are saved inside data field
        config = json.loads(
            json.dumps(saved_resource.config())
        )  # To deal with tuples and non-json types
        for k, v in history[0]["data"].items():
            if k == "client_port":
                # TODO seems like multiple tests in CI are saving the same resource at the same time, save needs
                # to return a version
                continue
            assert config[k] == v

    @pytest.mark.level("local")
    def test_sharing(
        self, saved_resource, friend_account_logged_in_docker_cluster_pk_ssh
    ):
        # Skip this test for ondemand clusters, because making
        # it compatible with ondemand_cluster requires changes
        # that break CI.
        # TODO: Remove this by doing some CI-specific logic.
        if saved_resource.__class__.__name__ == "OnDemandCluster":
            return
        if "Folder" in saved_resource.__class__.__name__:
            pytest.skip(
                "Skipping folder tests for now, since folders might be deprecated"
            )

        if TEST_ORG in saved_resource.rns_address:
            # Org resources require org membership or being granted explicit access - have a separate test for this
            return

        if saved_resource.rns_address.startswith("~"):
            # For `local_named_resource` resolve the rns address so it can be shared and loaded
            from runhouse.globals import rns_client

            saved_resource.rns_address = rns_client.local_to_remote_address(
                saved_resource.rns_address
            )

        saved_resource.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        # First try loading in same process/filesystem because it's more debuggable, but not as thorough
        resource_class_name = saved_resource.config().get("resource_type").capitalize()
        config = saved_resource.config()
        address = saved_resource.rns_address

        with friend_account():
            new_config = load_shared_resource_config(resource_class_name, address)
            if new_config["resource_subtype"] == "Secret":
                secret_name = new_config.pop("name")
                assert "loaded_secret_" in secret_name
                new_config["name"] = secret_name.replace("loaded_secret_", "")

            # Don't compare the client keys, their paths could differ locally and on the cluster
            # Cluster: '/home/runner/.ssh/sky-key
            # Locally: /home/runner/.rh/secrets/cluster-name/ssh-key
            new_config.get("data_config", {}).pop("client_keys", None)
            config.get("data_config", {}).pop("client_keys", None)

            assert new_config == config

        # TODO: If we are testing with an ondemand_cluster we to
        # sync sky key so loading ondemand_cluster from config works
        # Also need aws secret to load availability zones
        # secrets=["sky", "aws"],
        load_shared_resource_config_cluster = rh.function(
            load_shared_resource_config
        ).to(friend_account_logged_in_docker_cluster_pk_ssh)
        new_config = load_shared_resource_config_cluster(
            resource_class_name, saved_resource.rns_address
        )
        if new_config["resource_subtype"] == "Secret":
            secret_name = new_config.pop("name")
            assert "loaded_secret_" in secret_name
            new_config["name"] = secret_name.replace("loaded_secret_", "")

        new_config.get("data_config", {}).pop("client_keys", None)
        config.get("data_config", {}).pop("client_keys", None)
        assert new_config == config

    @pytest.mark.level("local")
    def test_sharing_org_resources(
        self, saved_resource, docker_cluster_pk_ssh_den_auth
    ):
        from tests.test_servers.conftest import summer

        # Skip this test for ondemand clusters, because making
        # it compatible with ondemand_cluster requires changes
        # that break CI.
        # TODO: Remove this by doing some CI-specific logic.
        if saved_resource.__class__.__name__ == "OnDemandCluster":
            return

        if TEST_ORG not in saved_resource.rns_address:
            # Skip if resource is not associated with an org
            return

        if saved_resource.rns_address.startswith("~"):
            # For `local_named_resource` resolve the rns address so it can be shared and loaded
            from runhouse.globals import rns_client

            saved_resource.rns_address = rns_client.local_to_remote_address(
                saved_resource.rns_address
            )

        saved_resource.share(
            users=["support@run.house"],
            access_level="read",
            notify_users=False,
        )

        # Org resources require org membership or being granted explicit access
        with pytest.raises(ValueError):
            with friend_account():
                load_shared_resource_config(
                    saved_resource.__class__.__name__,
                    saved_resource.rns_address,
                )

        resource_name = f"/{TEST_ORG}/summer_func"
        args = {"name": resource_name, "fn": summer}

        # Sending function to the cluster will save the function under the organization
        f = rh.function(**args).to(docker_cluster_pk_ssh_den_auth)
        init_args[id(f)] = args

        # Should be saved to Den under the org
        f.save()

        # Reload the function with the rns address pointing to the org
        # This should work given the requesting user has org access
        reloaded_func = rh.function(name=resource_name)
        assert (
            reloaded_func.system.rns_address
            == docker_cluster_pk_ssh_den_auth.rns_address
        )
        assert reloaded_func(1, 2) == 3

        with pytest.raises(Exception):
            # Friend account does not have org access, should not be able to reload the function
            # (unless also explicitly provided with resource access)
            with friend_account():
                rh.function(name=resource_name)

        # Friend account who is an org member should be able to reload the function without explicitly been given access
        with friend_account_in_org():
            rh.function(name=resource_name)

    # TODO API to run this on local_docker_slim when level == "local"
    @pytest.mark.skip
    def test_loading_in_new_fs(self, resource):
        pass



import runhouse as rh


def test_download_defaults():
    rh.globals.configs.defaults_cache["default_folder"] = "nonsense"
    local_defaults = rh.configs.load_defaults_from_file()
    local_defaults.pop("secrets")
    rh.configs.upload_defaults(defaults=local_defaults)
    loaded_defaults = rh.configs.download_defaults()
    assert local_defaults == loaded_defaults
    assert rh.globals.rns_client.default_folder == "nonsense"



import asyncio
import copy
import importlib
import inspect
import site
import sys
from concurrent.futures import ThreadPoolExecutor
from importlib import reload as importlib_reload
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

from apispec import APISpec
from pydantic import create_model

from runhouse.constants import DEFAULT_PROCESS_NAME
from runhouse.globals import obj_store, rns_client
from runhouse.logger import get_logger
from runhouse.resources.hardware import _current_cluster, _get_cluster_from, Cluster
from runhouse.resources.packages import Package
from runhouse.resources.resource import Resource
from runhouse.rns.utils.api import ResourceAccess, ResourceVisibility
from runhouse.servers.http import HTTPClient
from runhouse.servers.http.http_utils import CallParams
from runhouse.utils import (
    client_call_wrapper,
    generate_default_name,
    get_module_import_info,
    locate_working_dir,
)


# These are attributes that the Module's __getattribute__ logic should not intercept to run remotely
# and values that shouldn't be passed as state in put_resource
MODULE_ATTRS = [
    "_pointers",
    "_endpoint",
    "_client",
    "_visibility",
    "_name",
    "_rns_folder",
    "_system",
    "process",
    "dryrun",
    "_resolve",
    "_signature",
    "_dumb_signature_cache",
    "_openapi_spec",
    "access_level",
]

logger = get_logger(__name__)


class Module(Resource):
    RESOURCE_TYPE = "module"

    def __init__(
        self,
        pointers: Optional[Tuple] = None,
        signature: Optional[dict] = None,
        endpoint: Optional[str] = None,
        name: Optional[str] = None,
        system: Union[Cluster, str] = None,
        dryrun: bool = False,
        **kwargs,
    ):
        """
        Runhouse Module object.

        .. note::
                To create a Module, please use the factory method :func:`module`.
        """
        super().__init__(name=name, dryrun=dryrun, **kwargs)
        self._system = _get_cluster_from(
            system or _current_cluster(key="config"), dryrun=dryrun
        )
        self.process = obj_store.get_process_name() or DEFAULT_PROCESS_NAME
        is_builtin = hasattr(sys.modules["runhouse"], self.__class__.__qualname__)

        # If there are no pointers and this isn't a builtin module, we assume this is a user-created subclass
        # of rh.Module, and we need to do the factory constructor logic here.
        if not pointers and not is_builtin:

            # When creating a module as a subclass of rh.Module, we need to collect pointers here
            # If we're creating pointers, we're also local to the class definition and package, so it should be
            # set as the workdir (we can do this in a fancier way later)
            pointers = Module._extract_pointers(self.__class__)
        self._pointers = pointers
        self._endpoint = endpoint
        self._signature = signature
        self._dumb_signature_cache = None
        self._resolve = False
        self._openapi_spec = None

    def config(self, condensed: bool = True):
        config = super().config(condensed)
        if self.system:
            system = self._resource_string_for_subconfig(self.system, condensed)
        else:
            system = None

        config["system"] = system
        config["process"] = self.process
        if self._pointers:
            # For some reason sometimes this is coming back as a string, so we force it into a tuple
            config["pointers"] = tuple(self._pointers)

        # If not signature is set, we assume this is where the Module was created and we're local to its code.
        # We'll collect the signatures of all the methods here before saving or sending them somewhere.
        # Note that we even do this for built-in modules, because 1) we want their methods preserved in Den for when
        # they're called via HTTP and 2) we want to preserve the exact set of methods in case the methods on built-in
        # modules change across Runhouse versions.
        config["signature"] = self.signature(rich=True)

        # Only save the endpoint if it's present in _endpoint or externally accessible
        config["endpoint"] = self.endpoint(external=True)

        if self._openapi_spec:
            config["openapi_spec"] = self._openapi_spec

        return config

    @classmethod
    def from_config(
        cls, config: Dict, dryrun: bool = False, _resolve_children: bool = True
    ):
        if config.get("pointers"):
            config.pop("resource_subtype", None)
            logger.debug(f"Constructing module from pointers {config['pointers']}")
            (module_path, module_name, class_name) = config["pointers"]
            try:
                module_cls = cls._get_obj_from_pointers(
                    module_path, module_name, class_name
                )
                if not issubclass(module_cls, Module):
                    # Case when module was created through rh.module(new_class) factory, and needs to be
                    # made into a subclass of rh.Module. We'll follow the same flow as the subclass-created module
                    # below, where we don't call __init__ explicitly, because __init__ will call the subclass's init
                    # and this may a "type" module rather than an "instance". The user might instantiate it later, or
                    # it may be populated with attributes by the servlet's put_resource.
                    module_cls = _module_subclass_factory(
                        module_cls, config.get("pointers")
                    )
            except (ModuleNotFoundError, ImportError, AttributeError) as e:
                if (isinstance(e, ModuleNotFoundError) and module_name in str(e)) or (
                    (isinstance(e, ImportError) and class_name in str(e))
                    or (isinstance(e, AttributeError) and class_name in str(e))
                ):
                    system = config.get("system", None)
                    process = config.get("process", None)

                    # If we are on the same cluster where the module lives, we should be able to load the module from
                    # the pointers. So, we should just raise the exception if this is the case.
                    if system.on_this_cluster() and (
                        process == obj_store.servlet_name
                        and obj_store.has_local_storage
                    ):
                        # Could not load Module locally from within the system where it lives
                        raise e

                    # If we fail to construct the module class from the pointers, but system is elsewhere,
                    # we can still use this module from its signature alone.
                    module_cls = Module
                else:
                    # Other type of module or import error, such as dependency not installed
                    raise e

            # Module created as subclass of rh.Module may not have rh.Module's
            # constructor signature (e.g. system, process, etc.), so assign them manually
            # We don't call __init__ here because we don't know the signature of the subclass's __init__
            # If this resource was put on a cluster with put_resource, the servlet will be populating the rest
            # of the class-specific attributes.
            new_module = module_cls.__new__(module_cls)
            if _resolve_children:
                config = module_cls._check_for_child_configs(config)
            new_module.system = config.pop("system", None)
            new_module.process = config.pop("process", None)
            new_module.name = config.pop("name", None)
            new_module.access_level = config.pop("access_level", ResourceAccess.WRITE)
            new_module.visibility = config.pop("visibility", ResourceVisibility.PRIVATE)
            new_module._endpoint = config.pop("endpoint", None)
            new_module._pointers = config.pop("pointers", None)
            new_module._signature = config.pop("signature", None)
            new_module.dryrun = config.pop("dryrun", False)
            new_module._openapi_spec = config.pop("openapi_spec", None)
            # TODO not sure if we should keep this anymore given deserialization improvements
            new_module._resolve = config.pop("resolve", False)
            new_module._dumb_signature_cache = None
            return new_module

        if config.get("resource_subtype", None) == "module":
            config = Module._check_for_child_configs(config)
            return Module(**config, dryrun=dryrun)

        # If there are no class pointers, we assume the module is a built-in rh.Module subclass
        resource_class = getattr(
            sys.modules["runhouse"],
            config.pop("resource_subtype").capitalize(),
            None,
        )
        if not resource_class:
            raise TypeError(
                f"Could not find module associated with {config['resource_subtype']}"
            )
        config = resource_class._check_for_child_configs(config)
        return resource_class(**config, dryrun=dryrun)

    @classmethod
    def _check_for_child_configs(cls, config: dict):
        """Overload by child resources to load any resources they hold internally."""
        system = config.get("system")
        if isinstance(system, str) or isinstance(system, dict):
            config["system"] = _get_cluster_from(system)
        return config

    @property
    def system(self):
        """Cluster that the Module lives on."""
        return self._system

    @system.setter
    def system(self, new_system: Union[str, Cluster]):
        self._system = _get_cluster_from(new_system)

    def _compute_signature(self, rich=False):
        return {
            name: self.method_signature(method) if rich else None
            for (name, method) in inspect.getmembers(self.__class__)
            if not name[0] == "_"
            and name not in MODULE_ATTRS
            and name not in MODULE_METHODS
            and callable(method)
            # Checks if there's an arg called "local" in the method signature, and if so, if it's default is True.
            and not getattr(
                inspect.signature(method).parameters.get("local"), "default", False
            )
        }

    def signature(self, rich=False):
        if self._signature:
            return self._signature

        if not rich:
            if self._dumb_signature_cache is None:
                self._dumb_signature_cache = self._compute_signature(rich=False)
            return self._dumb_signature_cache

        self._signature = self._compute_signature(rich=True)
        return self._signature

    def method_signature(self, method):
        """Method signature, consisting of method properties to preserve when sending the method over the wire."""
        signature = inspect.signature(method)
        signature_metadata = {
            "signature": str(signature),
            "async": inspect.iscoroutinefunction(method)
            or inspect.isasyncgenfunction(method),
            "gen": inspect.isgeneratorfunction(method)
            or inspect.isasyncgenfunction(method),
            "doc": inspect.getdoc(method) if method.__doc__ else None,
            "annotations": str(inspect.getfullargspec(method).annotations),
        }

        return signature_metadata

    def _signature_extensions(self, item):
        """A way to manually extend the methods behind names in the signature. We can't use __getattr__
        because it gets too crazy and circular with the overloaded __getattribute__ below."""
        return None

    def endpoint(self, external: bool = False):
        """The endpoint of the module on the cluster. Returns an endpoint if one was manually set (e.g. if loaded
        down from a config). If not, request the endpoint from the Module's system.

        Args:
            external (bool, optional): If True and getting the endpoint from the system, only return an endpoint if
                it's externally accessible (i.e. not on localhost, not connected through as ssh tunnel). If False,
                return the endpoint even if it's not externally accessible. (Default: ``False``)
        """
        if self._endpoint:
            return self._endpoint

        if (
            self._system
            and hasattr(self._system, "endpoint")
            and self._system.endpoint(external=external)
        ):
            return f"{self._system.endpoint(external=external)}/{self.name}"

        return None

    def set_endpoint(self, new_endpoint: str):
        self._endpoint = new_endpoint

    def _client(self):
        """Return the client through which to interact with the source-of-truth Module. If this module is local,
        i.e. this module is its own source of truth, return None."""
        if (
            hasattr(self, "_system")
            and self._system
            and hasattr(self._system, "_client")
        ):
            return self._system._client()
        if (
            hasattr(self, "_endpoint")
            and self._endpoint
            and isinstance(self._endpoint, str)
        ):
            return HTTPClient.from_endpoint(
                self._endpoint, resource_address=self.rns_address
            )
        return None

    def _remote_init(self, *args, **kwargs):
        if len(self.__class__.__bases__) > 1:
            module_cls = self.__class__.__bases__[1]
        elif self._pointers:
            (module_path, module_name, class_name) = self._pointers
            # Reload needs to be false here, because if we reload the class, the reloaded class actually doesn't
            # match the class object of which self was a subclass, so we can't call super().__init__ on it within
            # module_cls's __init__.
            # We'd get "TypeError: super(type, obj): obj must be an instance or subtype of type"
            module_cls = self._get_obj_from_pointers(
                module_path, module_name, class_name, reload=False
            )
        else:
            module_cls = self.__class__

        # Change around the MRO so that the module_cls is the first parent class, and Module is the second,
        # so methods like .to default to the module_cls and not Module's when on the cluster.
        # This is a small price to pay for matching PyTorch's .to API. If it creates too much craziness we can
        # revisit.
        # class NewSubclass(module_cls, Module):
        #     pass
        #
        # self.__class__ = NewSubclass

        module_cls.__init__(self, *args, **kwargs)

    @staticmethod
    def _get_obj_from_pointers(module_path, module_name, obj_name, reload=True):
        """Helper method to load a class or function from a module path, module name, and class name."""
        if module_path:
            abs_path = str((Path.home() / module_path).expanduser().resolve())
            if not abs_path:
                logger.debug(f"Could not find module path {module_path}")
            elif abs_path not in sys.path:
                sys.path.insert(0, abs_path)
                logger.debug(f"Appending {module_path} to sys.path")

        # This updates the sys.path with any new paths that have been added since the last time we imported
        # e.g. if the user ran cluster.run_bash(["pip install my_package"]) since this module was created.
        importlib_reload(site)

        if module_name in obj_store.imported_modules and reload:
            importlib.invalidate_caches()
            obj_store.imported_modules[module_name] = importlib.reload(
                obj_store.imported_modules[module_name]
            )
            logger.debug(f"Reloaded module {module_name}")
        else:
            logger.debug(f"Importing module {module_name}")
            obj_store.imported_modules[module_name] = importlib.import_module(
                module_name
            )

        return getattr(obj_store.imported_modules[module_name], obj_name)

    def __getstate__(self):
        # Exclude anything already being sent in the config and private module attributes
        state = {}
        # We only send over state for instances, not classes
        if not isinstance(self, type):
            state = {
                attr: val
                for attr, val in self.__dict__.items()
                if attr not in MODULE_ATTRS and attr not in MODULE_METHODS
            }
        return state

    def default_name(self):
        return (self._pointers[2] if self._pointers else None) or generate_default_name(
            prefix=self.__class__.__qualname__.lower()
        )

    def to(
        self,
        system: Union[str, Cluster],
        process: Optional[Union[str, Dict]] = None,
        name: Optional[str] = None,
        sync_local: bool = True,
    ):
        """Send the module to a specified process on a cluster. This will sync over relevant code for the module
        to run on the cluster, and return a remote_module object that will wrap remote calls to the module
        living on the cluster.

        Args:
            system (str or Cluster): The cluster to setup the module and process on.
            process (str or Dict, optional): The process to run the module on. If it's a Dict, it will be explicitly
                created with those args. or the set of requirements necessary to run the module. If no process is
                specified, the module will be sent to the default_process created when the cluster is created
                (Default: ``None``)
            name (Optional[str], optional): Name to give to the module resource, if you wish to rename it.
                (Default: ``None``)
            sync_local (bool, optional): Whether to sync up and use the local module on the cluster. If ``False``,
                don't sync up and use the equivalent module found on the cluster. (Default: ``True``)

        Example:
            >>> local_module = rh.module(my_class)
            >>> cluster_module = local_module.to("my_cluster")
        """
        if system == self.system:
            if name and not self.name == name:
                # TODO return duplicate object under new name, don't rename
                self.rename(name)
            return self

        if system == "here":
            from runhouse import here

            system = here

        if system == "file":
            raise ValueError(
                "Need an initialized local server in order to put a module onto `rh.here`. Please run `runhouse server restart` first on your local machine."
            )

        system = (
            _get_cluster_from(system, dryrun=self.dryrun) if system else self.system
        )

        if not system:
            raise ValueError("No system specified to send module to.")

        new_name = name or self.name or self.default_name()
        if self.rns_address:
            new_name = f"{self._rns_folder}/{new_name}"

        if process is None:
            # Make this an empty dict so it'll be picked up for the process creation
            process = {}

        # If process wasn't specified and name wasn't specified, we're going to use a default name tied to this
        # module's class name and kill the old process if it exists.
        if isinstance(process, Dict):
            if "name" not in process:
                ephemeral_process_name = f"{new_name}_process"
                if ephemeral_process_name in system.list_processes():
                    system.kill_process(ephemeral_process_name)

                # Now we make sure that the process created with the args provided by the user
                process["name"] = ephemeral_process_name

            process = system.ensure_process_created(**process)
        else:
            # If name was specified, we don't delete and overwrite, we just let it be
            system.ensure_process_created(name=process)

        if not isinstance(process, str):
            raise ValueError(
                "Process arg must be a string name of the process or a dict of create_process kwargs."
            )

        # We need to change the pointers to the remote import path if we're sending this module to a remote cluster,
        # and we need to add the local path to the module to the requirements if it's not already there.
        remote_import_path = None
        if sync_local and (self._pointers or getattr(self, "fn_pointers", None)):
            pointers = self._pointers if self._pointers else self.fn_pointers

            # Update the system reqs with the local path to the module if it's not already there
            (
                local_path_containing_module,
                should_add,
            ) = Module._get_local_path_containing_module(pointers[0], system.reqs)
            if should_add:
                system.reqs = [str(local_path_containing_module)] + system.reqs

            system.install_package(str(local_path_containing_module))

            # Figure out what the import path would be on the remote system
            remote_import_path = str(
                local_path_containing_module.name
                / Path(pointers[0]).relative_to(local_path_containing_module)
            )

        # We need to backup the system here so the __getstate__ method of the cluster
        # doesn't wipe the client of this function's cluster when deepcopy copies it.
        hw_backup = self.system
        self.system = None
        new_module = copy.copy(self)
        self.system = hw_backup

        new_module.system = system
        new_module.process = process
        new_module.dryrun = True

        # Set remote import path
        if remote_import_path:
            if new_module._pointers:
                new_module._pointers = (
                    remote_import_path,
                    new_module._pointers[1],
                    new_module._pointers[2],
                )
            else:
                new_module.fn_pointers = (
                    remote_import_path,
                    new_module.fn_pointers[1],
                    new_module.fn_pointers[2],
                )

        if isinstance(system, Cluster):
            new_module.name = new_name
            # TODO dedup with _extract_state
            # Exclude anything already being sent in the config and private module attributes
            excluded_state_keys = list(new_module.config().keys()) + MODULE_ATTRS
            state = {}
            # We only send over state for instances, not classes
            if not isinstance(self, type):
                state = {
                    attr: val
                    for attr, val in self.__dict__.items()
                    if attr not in excluded_state_keys
                }
            logger.info(
                f"Sending module {new_module.name} of type {type(new_module)} to {system.name or 'local Runhouse daemon'}"
            )
            system.put_resource(new_module, state, dryrun=True)

        if rns_client.autosave_resources():
            new_module.save()

        return new_module

    def get_or_to(
        self,
        system: Union[str, Cluster],
        process: Optional[Union[str, Dict]] = None,
        name: Optional[str] = None,
    ):
        """Check if the module already exists on the cluster, and if so return the module object.
        If not, put the module on the cluster and return the remote module.

        Args:
            system (str or Cluster): The system to setup the module.
            process (str or Dict, optional): The process to run the module on, if it's a Dict, it will be explicitly
                created with those args. or the set of requirements necessary to run the module. (Default: ``None``)
            name (Optional[str], optional): Name to give to the module resource, if you wish to rename it.
                (Default: ``None``)

        Example:
            >>> remote_df = Model().get_or_to(my_cluster, name="remote_model")
        """
        name = self.name or name
        if not name:
            raise ValueError(
                "You must specify a name for the module if you want to get_or_to it."
            )
        system = _get_cluster_from(system)
        try:
            remote = system.get(name, remote=True)
        except KeyError:
            remote = None
        if remote:
            return remote
        self.name = name
        return self.to(system, process=process)

    def __getattribute__(self, item):
        """Override to allow for remote execution if system is a remote cluster. If not, the subclass's own
        __getattr__ will be called."""
        if (
            item in MODULE_METHODS
            or item in MODULE_ATTRS
            or not hasattr(self, "_client")
        ):
            return super().__getattribute__(item)

        try:
            name = super().__getattribute__("_name")
        except AttributeError:
            return super().__getattribute__(item)

        if item not in self.signature(rich=False) or not name:
            return super().__getattribute__(item)

        # Do this after the signature and name check because it's potentially expensive
        client = super().__getattribute__("_client")()

        if not client:
            return super().__getattribute__(item)

        system = super().__getattribute__("_system")

        is_coroutine_function = (
            self.signature(rich=True)[item]["async"]
            and not self.signature(rich=True)[item]["gen"]
        )

        class RemoteMethodWrapper:
            """Helper class to allow methods to be called with __call__, remote, or run."""

            def __call__(self, *args, **kwargs):
                # stream_logs and run_name are both supported args here, but we can't include them explicitly because
                # the local code path here will throw an error if they are included and not supported in the
                # method signature.

                # Always take the user overrided behavior if they want it
                if "run_async" in kwargs:
                    run_async = kwargs.pop("run_async")
                else:
                    run_async = is_coroutine_function

                if run_async:
                    return client_call_wrapper(
                        client,
                        system,
                        "acall",
                        name,
                        item,
                        run_name=kwargs.pop("run_name", None),
                        stream_logs=kwargs.pop("stream_logs", True),
                        remote=kwargs.pop("remote", False),
                        data={"args": args, "kwargs": kwargs},
                    )
                else:
                    return client_call_wrapper(
                        client,
                        system,
                        "call",
                        name,
                        item,
                        run_name=kwargs.pop("run_name", None),
                        stream_logs=kwargs.pop("stream_logs", True),
                        remote=kwargs.pop("remote", False),
                        data={"args": args, "kwargs": kwargs},
                    )

            def remote(self, *args, stream_logs=True, run_name=None, **kwargs):
                return self.__call__(
                    *args,
                    stream_logs=stream_logs,
                    run_name=run_name,
                    remote=True,
                    **kwargs,
                )

            def run(self, *args, stream_logs=False, run_name=None, **kwargs):
                return self.__call__(
                    *args,
                    stream_logs=stream_logs,
                    run_name=run_name,
                    **kwargs,
                )

        return RemoteMethodWrapper()

    # Overload the reduce method so all subclasses of Module can be pickled as
    # a Module object, and not as a subclass of Module (which could cause deserialization
    # errors when pickle attempts to look up those classes remotely). Now the "envelope"
    # for how these objects are pickled is always the Module config, which means we can
    # always deserialize them as Module objects (as long as we can find the pointers remotely).
    def __reduce__(self):
        return Resource.from_config, (self.config(),), self.__getstate__()

    def refresh(self):
        """Update the resource in the object store."""
        if not self.system or not self.name:
            return self
        if self._system.on_this_cluster():
            return obj_store.get(self._name)
        elif isinstance(self._system, Cluster):
            return self._system.get(self._name, remote=True)
        else:
            return self

    def replicate(
        self,
        num_replicas: int = 1,
        replicas_per_node: Optional[int] = None,
        names: List[str] = None,
        processes: Optional[List["Process"]] = None,
        parallel: bool = False,
    ):
        """Replicate the module on the cluster in new processes and return the new modules.

        Args:
            num_replicas (int, optional): Number of replicas of the module to create. (Default: 1)
            replicas_per_node (int, optional): The number of replicas to create per node. (Defualt: ``None``)
            names (List[str], optional): List for the names for the replicas, if specified. (Default: ``None``)
            processes (List[Process], optional): List of the processes for the replicas, if specified. (Default: ``None``)
            parallel (bool, optional): Whether to create the replicas in parallel. (Default: ``False``)
        """
        if not self.system or not self.name:
            raise ValueError(
                "Cannot replicate a module that is not on a cluster. Please send the module to a cluster first."
            )
        if not isinstance(self.system, Cluster):
            raise ValueError(
                "Cannot replicate a module that is not on a cluster. Please send the module to a cluster first."
            )
        if processes and not len(processes) == num_replicas:
            raise ValueError(
                "If processes is a list, it must be the same length as num_replicas."
            )
        if names and not len(names) == num_replicas:
            raise ValueError(
                "If names is a list, it must be the same length as num_replicas."
            )
        if not processes and not self.process:
            raise ValueError(
                "Cannot replicate without a process. Please send the module or function to a process on the system first."
            )

        def create_replica(i):
            name = names[i] if isinstance(names, list) else f"{self.name}_replica_{i}"

            if isinstance(processes, list):
                replica_process_name = processes[i]
            else:
                replica_process_name = f"{self.process}_replica_{i}"

                # The replica process should have the same requirements as the original process
                new_create_process_params = self.system.list_processes()[self.process]

                # Change the arguments for the new process to reflect the replica
                new_create_process_params["name"] = replica_process_name
                if replicas_per_node is not None:
                    if new_create_process_params["compute"]:
                        raise ValueError(
                            "Cannot specify replicas_per_node if other compute requirements are specified."
                        )
                    new_create_process_params["compute"] = {
                        "node_idx": i // replicas_per_node
                    }
                self.system.ensure_process_created(**new_create_process_params)

            new_module = copy.copy(self)
            new_module.local.name = name
            new_module.local.process = None
            new_module.local.system = None

            new_module = new_module.to(self.system, process=replica_process_name)
            return new_module

        if parallel:
            from multiprocessing import pool

            with pool.ThreadPool() as p:
                return list(p.imap(create_replica, range(num_replicas)))

        return [create_replica(i) for i in range(num_replicas)]

    def distribute(
        self,
        distribution: str,
        name: Optional[str] = None,
        num_replicas: Optional[int] = None,
        replicas_per_node: Optional[int] = None,
        replication_kwargs: Dict = {},
        **distribution_kwargs,
    ):
        """Distribute the module on the cluster and return the distributed module.

        Args:
            distribution (str): The distribution method to use, e.g. "pool", "ray", "pytorch", or "tensorflow".
            name (str, optional): The name to give to the distributed module, if applicable. Overwrites current module
                name by default. (Default: ``None``)
            num_replicas (int, optional): The number of replicas to create. (Default: ``None``)
            replicas_per_node (int, optional): The number of replicas to create per node. (Default: ``None``)
            replication_kwargs (Dict): The keyword arguments to pass to the replicate method. (Default: {})
            distribution_kwargs: The keyword arguments to pass to the distribution method.
        """

        if not self.system or not self.name:
            raise ValueError(
                "Cannot distribute a module that is not on a cluster. Please send the module to a cluster first."
            )

        if not self.system.on_this_cluster():
            # Distribute the module on the cluster and return the distributed module as a stub.
            return self.system.call(
                self.name,
                "distribute",
                distribution,
                name,
                num_replicas,
                replicas_per_node,
                replication_kwargs,
                remote=True,
                **distribution_kwargs,
            )

        if distribution == "pool":
            from runhouse.resources.distributed.distributed_pool import DistributedPool

            if not num_replicas:
                raise ValueError(
                    "``num_replicas`` argument must be provided for distribution type pool"
                )

            replicas = self.replicate(
                num_replicas=num_replicas,
                replicas_per_node=replicas_per_node,
                **replication_kwargs,
            )
            name = name or f"distributed_{self.name}"
            pooled_module = DistributedPool(
                **distribution_kwargs, name=name, replicas=replicas
            ).to(self.system, process=self.process)
            return pooled_module
        elif distribution == "ray":
            from runhouse.resources.distributed.ray_distributed import RayDistributed

            name = name or f"ray_{self.name}"
            ray_module = RayDistributed(
                **distribution_kwargs, name=name, module=self
            ).to(system=self.system, process=self.process)
            return ray_module
        elif distribution == "dask":
            from runhouse.resources.distributed.dask_distributed import DaskDistributed

            name = name or f"dask_{self.name}"
            dask_module = DaskDistributed(
                name=name,
                module=self,
                **distribution_kwargs,
            ).to(system=self.system, process=self.process)
            return dask_module
        elif distribution == "pytorch":
            from runhouse.resources.distributed.pytorch_distributed import (
                PyTorchDistributed,
            )

            if not num_replicas:
                num_nodes = len(self.system.ips)
                if not replicas_per_node:
                    replicas_per_node = self.system._gpus_per_node() or 1
                num_replicas = replicas_per_node * num_nodes

            replicas = self.replicate(
                num_replicas=num_replicas,
                replicas_per_node=replicas_per_node,
                **replication_kwargs,
            )
            name = name or f"pytorch_{self.local.name}"
            ptd_module = PyTorchDistributed(
                **distribution_kwargs, name=name, replicas=replicas
            ).to(system=self.system, process=self.process)
            return ptd_module

    @property
    def remote(self):
        """Helper property to allow for access to remote properties, both public and private. Returning functions
        is not advised.

        Example:
            >>> my_module.remote.my_property
            >>> my_module.remote._my_private_property
            >>> my_module.remote.size = 14
        """
        client = super().__getattribute__("_client")()
        system = super().__getattribute__("_system")
        name = super().__getattribute__("_name")

        outer_super_gettattr = super().__getattribute__
        outer_super_setattr = super().__setattr__

        class RemotePropertyWrapper:
            @classmethod
            def __getattribute__(cls, item):
                # TODO[DG] revise
                if not client or not name:
                    return outer_super_gettattr(item)

                return client_call_wrapper(
                    client, system, "call", name, item, stream_logs=False
                )

            @classmethod
            def __setattr__(cls, key, value):
                if not client or not name:
                    return outer_super_setattr(key, value)

                return client_call_wrapper(
                    client,
                    system,
                    "call",
                    key=name,
                    method_name=key,
                    data={"args": [value], "kwargs": {}},
                )

            @classmethod
            def __call__(cls, *args, **kwargs):
                return system.get(name, *args, **kwargs)

        return RemotePropertyWrapper()

    @property
    def local(self):
        """Helper property to allow for access to local properties, both public and private.

        Example:
            >>> my_module.local.my_property
            >>> my_module.local._my_private_property

            >>> my_module.local.size = 14
        """
        outer_super_gettattr = super().__getattribute__
        outer_super_setattr = super().__setattr__

        class LocalPropertyWrapper:
            @classmethod
            def __getattribute__(cls, item):
                return outer_super_gettattr(item)

            @classmethod
            def __setattr__(cls, key, value):
                return outer_super_setattr(key, value)

        return LocalPropertyWrapper()

    def fetch(self, item: str = None, **kwargs):
        """Helper method to allow for access to remote state, both public and private. Fetching functions
        is not advised. `system.get(module.name).resolved_state()` is roughly equivalent to `module.fetch()`.

        Example:
            >>> my_module.fetch("my_property")
            >>> my_module.fetch("my_private_property")

            >>> MyRemoteClass = rh.module(my_class).to(system)
            >>> MyRemoteClass(*args).fetch() # Returns a my_class instance, populated with the remote state

            >>> my_module.fetch() # Returns the data of the blob, due to overloaded ``resolved_state`` method

            >>> class MyModule(rh.Module):
            >>>     # ...
            >>>
            >>> MyModule(*args).to(system).fetch() # Returns the full remote module, including private and public state
        """
        name = super().__getattribute__("_name")

        client = super().__getattribute__("_client")()
        system = super().__getattribute__("_system")

        if item is not None:
            if not client or not name:
                return super().__getattribute__(item)
            return client_call_wrapper(client, system, "call", name, item)
        else:
            if not client or not name:
                return self.resolved_state(**kwargs)
            return client_call_wrapper(
                client,
                system,
                "call",
                name,
                "resolved_state",
                data={"args": [], "kwargs": kwargs},
            )

    async def fetch_async(
        self, key: str, remote: bool = False, stream_logs: bool = False
    ):
        """Async version of fetch. Can't be a property like `fetch` because __getattr__ can't be awaited.

        Example:
            >>> await my_module.fetch_async("my_property")
            >>> await my_module.fetch_async("_my_private_property")
        """
        client = super().__getattribute__("_client")()
        system = super().__getattribute__("_system")
        name = super().__getattribute__("_name")

        def call_wrapper():
            if not key:
                return client_call_wrapper(client, system, "get", name, remote=remote)

            if isinstance(system, Cluster) and name and system.on_this_cluster():
                obj_store_obj = obj_store.get(name, default=None)
                if obj_store_obj:
                    return obj_store_obj.__getattribute__(key)
                else:
                    return self.__getattribute__(key)
            return client_call_wrapper(client, system, "call", name, key, remote=remote)

        try:
            is_gen = (key and hasattr(self, key)) and inspect.isasyncgenfunction(
                super().__getattribute__(key)
            )
        except AttributeError:
            is_gen = self.signature(rich=False).get(key, {}).get("gen", False)

        if is_gen:

            async def async_gen():
                for res in call_wrapper():
                    yield res

            return async_gen()

        _executor = ThreadPoolExecutor(1)

        async def async_call():
            return await loop.run_in_executor(_executor, call_wrapper)

        loop = asyncio.get_event_loop()
        return await asyncio.create_task(async_call())

    async def set_async(self, key: str, value):
        """Async version of property setter.

        Example:
            >>> await my_module.set_async("my_property", my_value)
            >>> await my_module.set_async("_my_private_property", my_value)
        """
        client = super().__getattribute__("_client")()
        if not client or not self._name:
            return super().__setattr__(key, value)

        def call_wrapper():
            return self._client().call(
                key=self._name,
                method_name=key,
                data={"args": [value], "kwargs": {}},
                stream_logs=False,
            )

        _executor = ThreadPoolExecutor(1)

        async def async_call():
            return await loop.run_in_executor(_executor, call_wrapper)

        loop = asyncio.get_event_loop()
        return await asyncio.create_task(async_call())

    def resolve(self):
        """Specify that the module should resolve to a particular state when passed into a remote method. This is
        useful if you want to revert the module's state to some "Runhouse-free" state once it is passed into a
        Runhouse-unaware function. For example, if you call a Runhouse-unaware function with ``.remote()``,
        you will be returned a module which wraps your data. If you want to pass that module into another function
        that operates on the original data (e.g. a function that takes a numpy array), you can call
        ``my_second_fn(my_func.resolve())``, and ``my_func`` will be replaced with the contents of its ``.data`` on the
        cluster before being passed into ``my_second_fn``.

        Resolved state is defined by the ``resolved_state`` method. By default, modules created with the
        ``rh.module`` factory constructor will be resolved to their original non-module-wrapped class (or best attempt).
        Modules which are defined as a subclass of ``Module`` will be returned as-is, as they have no other
        "original class."

        Example:
            >>> my_module = rh.module(my_class)
            >>> my_remote_fn(my_module.resolve()) # my_module will be replaced with the original class `my_class`

            >>> my_result_module = my_remote_fn.call.remote(args)
            >>> my_other_remote_fn(my_result_module.resolve()) # my_result_module will be replaced with its data

        """
        self._resolve = True
        return self

    def resolved_state(self):
        """Return the resolved state of the module. By default, this is the original class of the module if it was
        created with the ``module`` factory constructor."""
        if not self._pointers:
            self._resolve = False
            return self

        (module_path, module_name, class_name) = self._pointers
        original_class = self._get_obj_from_pointers(
            module_path, module_name, class_name
        )
        if issubclass(original_class, Module):
            self._resolve = False
            return self

        if not self.__dict__:
            # This is a non-instantiated Module, i.e. represents a class rather than an instance
            return original_class

        new_module = original_class.__new__(original_class)
        # TODO pop out any attributes that are not in the original class?
        new_module.__dict__ = self.__dict__
        return new_module

    def _save_sub_resources(self, folder: str = None):
        if isinstance(self.system, Resource) and self.system.name:
            self.system.save(folder=folder)

    def rename(self, name: str):
        """Rename the module.

        Args:
            name (str): Name to rename the module to.
        """
        if self.name == name or self.rns_address == name:
            return
        old_name = self.name
        self.name = name  # Goes through Resource setter to parse name properly (e.g. if rns path)
        if (
            self.system
            and isinstance(self.system, Cluster)
            and self.system.on_this_cluster()
        ):
            # We rename the object with the full rns_address here because if it's a resource, we want the
            # object stored in the obj store to have the updated rns_address, not just the updated key.
            obj_store.rename(old_key=old_name, new_key=self.rns_address or self.name)
        elif self._client():
            self._client().rename(
                old_key=old_name, new_key=self.rns_address or self.name
            )

    def save(self, name: str = None, overwrite: bool = True, folder: str = None):
        # Need to override Resource's save to handle key changes in the obj store
        # Also check that this is a Module and not a File

        # Make sure to generate the openapi spec before saving
        _ = self.openapi_spec()

        old_rns_address = self.rns_address
        if name:
            _, base_name = rns_client.split_rns_name_and_path(
                rns_client.resolve_rns_path(name)
            )
            if self.name != base_name:
                if overwrite:
                    self.rename(name)
                else:
                    self.name = name
                    if isinstance(self.system, Cluster):
                        self.system.put_resource(self)

        res = super().save(overwrite=overwrite, folder=folder)

        if old_rns_address != self.rns_address:
            self.remote.name = self.rns_address

        return res

    def share(self, *args, visibility=None, **kwargs):
        if visibility and not visibility == self.visibility:
            self.visibility = visibility
            self.remote.visibility = (
                visibility  # Sets the visibility on the remote resource
            )
        return super().share(*args, **kwargs, visibility=visibility)

    @staticmethod
    def _is_running_in_notebook(module_path: Union[str, None]) -> bool:
        """Returns True if running in a notebook, False otherwise"""

        # Check if running in an IPython notebook
        # TODO better way of detecting if in a notebook or interactive Python env
        if not module_path or module_path.endswith("ipynb"):
            return True

        # Check if running in a marimo notebook
        try:
            import marimo as mo

            return mo.running_in_notebook()
        except (ImportError, ModuleNotFoundError):
            # marimo not installed
            return False

    @staticmethod
    def _extract_pointers(raw_cls_or_fn: Union[Type, Callable]):
        """Get the path to the module, module name, and function name to be able to import it on the server"""
        if not (isinstance(raw_cls_or_fn, Type) or isinstance(raw_cls_or_fn, Callable)):
            raise TypeError(
                f"Expected Type or Callable but received {type(raw_cls_or_fn)}"
            )

        root_path, module_name, cls_or_fn_name = get_module_import_info(raw_cls_or_fn)

        return (
            root_path,
            module_name,
            cls_or_fn_name,
        )

    @staticmethod
    def _get_local_path_containing_module(
        root_path: str, reqs: List[str]
    ) -> Tuple[Path, bool]:
        """
        Find the directory containing the module root path in the reqs list.

        If it is not found, find the working directory that contains the module root path and return that,
        along with a flag indicating whether the module should be added to the reqs list.

        """

        # First, check if the module is already included in one of the directories in reqs
        local_path_containing_module = None
        for req in reqs:
            if isinstance(req, str):
                req = Package.from_string(req)

            if (
                isinstance(req, Package)
                and not isinstance(req.install_target, str)
                # and req.install_target.is_local()
            ):
                local_path_containing_module = Path(req.install_target.local_path)

            if local_path_containing_module:
                try:
                    # Module path relative to package
                    Path(root_path).relative_to(local_path_containing_module)
                    break
                except ValueError:  # Not a subdirectory
                    local_path_containing_module = None
                    pass

        # Only add this to the system's reqs if the module is not in one of the directories in reqs
        add = local_path_containing_module is None

        if not local_path_containing_module:
            # If the module is not in one of the directories in reqs, we just use the full path,
            # then we'll create a new "req" containing the Module
            # TODO: should this "req" be a Package? I'll just start with a string for now
            local_path_containing_module = Path(locate_working_dir(root_path))

        return local_path_containing_module, add

    def openapi_spec(self, spec_name: Optional[str] = None):
        """Generate an OpenAPI spec for the module.

        Args:
            spec_name (str, optional): Spec name for the OpenAPI spec.
        """

        """ TODO: This breaks if the module has type annotations that are classes, and not standard library or
        typing types.

        Maybe we can do something using: https://github.com/kuimono/openapi-schema-pydantic to allow
        nested Pydantic models easily as schemas?

        TODO: What happens if there is an empty function, will this work with an empty body even though it is
        marked as required?
        """
        if self._openapi_spec is not None:
            return self._openapi_spec

        spec_name = spec_name or self.name

        if not spec_name:
            raise ValueError(
                "Module must have a name to generate an OpenAPI spec. Try saving the module first."
            )

        spec = APISpec(
            title=spec_name,
            version="1.0.0",
            openapi_version="3.0.0",
            plugins=[],
        )

        base_module_class_methods = {
            m[0] for m in inspect.getmembers(Module, predicate=inspect.isfunction)
        }
        for method_name, method in inspect.getmembers(
            self.__class__, predicate=inspect.isfunction
        ):
            if method_name.startswith("_") or method_name in base_module_class_methods:
                continue

            spec.path(
                path=f"/{spec_name}/{method_name}",
                operations={
                    "post": {
                        "summary": method.__doc__.strip().split("\n")[0]
                        if method.__doc__
                        else "",
                        "description": method.__doc__.strip() if method.__doc__ else "",
                        "requestBody": {
                            "content": {
                                "application/json": {
                                    "schema": f"{method_name}_body_schema",
                                },
                            },
                            "required": True,
                        },
                        "responses": {
                            "200": {
                                "description": "A response containing an output_type, data, and serialization, and optionally an error and traceback."
                            },
                        },
                    },
                },
            )

            # Generate schema for the Request Body
            params = {}
            for param_name, param in inspect.signature(method).parameters.items():
                if param_name == "self":
                    continue

                if param.annotation and param.annotation != inspect.Parameter.empty:
                    param_type = (
                        param.annotation.__name__
                        if isinstance(param.annotation, type)
                        else param.annotation
                    )
                else:
                    param_type = str

                params[param_name] = (
                    param_type,
                    param.default if param.default != inspect.Parameter.empty else ...,
                )

            module_method_params = create_model(
                f"{method_name}_schema", **params
            ).model_json_schema()
            module_method_params["title"] = "kwargs"

            request_body_schema = CallParams.model_json_schema()
            request_body_schema["properties"]["data"] = {
                "title": "data",
                "type": "object",
                "properties": {
                    "kwargs": module_method_params,
                },
                "required": ["kwargs"],
            }
            request_body_schema["title"] = f"{method_name} Params"
            spec.components.schema(f"{method_name}_body_schema", request_body_schema)

        self._openapi_spec = spec.to_dict()
        return self._openapi_spec

    # Found in python decorator logic, maybe use
    # func_name = getattr(f, '__qualname__', f.__name__)
    # module_name = getattr(f, '__module__', '')
    # if module_name:
    #     full_name = f'{module_name}.{func_name}'
    # else:
    #     full_name = func_name


def _module_subclass_factory(cls, cls_pointers):
    def __init__(
        self,
        *args,
        system=None,
        dryrun=False,
        pointers=cls_pointers,
        signature=None,
        name=None,
        **kwargs,
    ):
        # args and kwargs are passed to the cls's __init__ method if this is being called on a cluster. They
        # shouldn't be passed otherwise.
        Module.__init__(
            self,
            pointers=pointers,
            signature=signature,
            name=name,
            system=system,
            dryrun=dryrun,
        )
        # This allows a class which is already on the cluster to construct an instance of itself with a factory
        # method, e.g. my_module = MyModuleCls.factory_constructor(*args, **kwargs)
        if self.system and self.system.on_this_cluster():
            self._remote_init(*args, **kwargs)

    @classmethod
    def _module_init_only(cls, *args, **kwargs):
        new_module = cls.__new__(cls)
        Module.__init__(new_module, *args, **kwargs)
        return new_module

    def __call__(
        self,
        *args,
        dryrun=False,
        name=None,
        **kwargs,
    ):
        system_backup = self.system
        new_module = copy.copy(self)
        new_module.system = system_backup
        # Create a copy of the item on the cluster under the new name
        new_module.name = name or self.name
        new_module.dryrun = dryrun
        process = kwargs.get("process") or self.process
        if not new_module.dryrun and new_module.system:
            # We use system.put_resource here because the signatures for HTTPClient.put_resource and
            # obj_store.put_resource are different, but we should fix that.

            # If the system is still a string, we know that the user has _no_ access to the Cluster
            # If they were able to discover the cluster, their system string would be replaced by a Cluster object
            # when _check_for_child_configs --> _get_cluster_from was called
            if isinstance(new_module.system, str):
                raise ValueError(
                    "You must have access to the underlying cluster for this unconstructed Module in order to put a resource on it."
                )
            new_module.system.put_resource(new_module, process=process)
            new_module.system.call(new_module.name, "_remote_init", *args, **kwargs)
        else:
            new_module._remote_init(*args, **kwargs)

        return new_module

    methods = {
        "__init__": __init__,
        "__call__": __call__,
        "_module_init_only": _module_init_only,
    }
    new_type = type(cls_pointers[2], (Module, cls), methods)
    return new_type


def module(
    cls: [Type] = None,
    name: Optional[str] = None,
    load_from_den: bool = True,
    dryrun: bool = False,
):
    """Returns a Module object, which can be used to instantiate and interact with the class remotely.

    The behavior of Modules (and subclasses thereof) is as follows:
        - Any callable public method of the module is intercepted and executed remotely over rpc, with exception of
          certain functions Python doesn't make interceptable (e.g. __call__, __init__), and methods of the Module
          class (e.g. ``to``, ``fetch``, etc.). Properties and private methods are not intercepted, and will be
          executed locally.
        - Any method which executes remotely may be called normally, e.g. ``model.forward(x)``, or asynchronously,
          e.g. ``key = model.forward.run(x)`` (which returns a key to retrieve the result with
          ``cluster.get(key)``), or with ``run_obj = model.train.remote(x)``, which runs synchronously but returns
          a remote object to avoid passing heavy results back over the network.
        - Setting attributes, both public and private, will be executed remotely, with the new values only being
          set in the remote module and not the local one. This excludes any methods or attribtes of the Module class
          proper (e.g. ``system`` or ``name``), which will be set locally.
        - Attributes, private properties can be fetched with the ``remote`` property, and the full resource can be
          fetched using ``.fetch()``, e.g. ``model.remote.weights``, ``model.remote.__dict__``, ``model.fetch()``.
        - When a module is sent to a cluster, it's public attribtes are serialized, sent over, and repopulated in the
          remote instance. This means that any changes to the module's attributes will not be reflected in the remote


    Args:
        cls: The class to instantiate.
        name (Optional[str], optional): Name to give the module object, to be reused later on. (Default: ``None``)
        load_from_den (bool, optional): Whether to try loading the module from Den. (Default: ``True``)
        dryrun (bool, optional): Whether to create the Module if it doesn't exist, or load a Module object as a dryrun.
            (Default: ``False``)

    Returns:
        Module: The resulting module.

    Example - creating a module by defining an rh.Module subclass:
        >>> import runhouse as rh
        >>> import transformers
        >>>
        >>> # Sample rh.Module class
        >>> class Model(rh.Module):
        >>>    def __init__(self, model_id, device="cpu"):
        >>>        # Note that the code here will be run in your local environment prior to being sent to
        >>>        # to a cluster. For loading large models/datasets that are only meant to be used remotely,
        >>>        # we recommend using lazy initialization (see tokenizer and model attributes below).
        >>>        super().__init__()
        >>>        self.model_id = model_id
        >>>        self.device = device
        >>>
        >>>    @property
        >>>    def tokenizer(self):
        >>>        # Lazily initialize the tokenizer remotely only when it is needed
        >>>        if not hasattr(self, '_tokenizer'):
        >>>            self._tokenizer = transformers.AutoTokenizer.from_pretrained(self.model_id)
        >>>        return self._tokenizer
        >>>
        >>>    @property
        >>>    def model(self):
        >>>        if not hasattr(self, '_model'):
        >>>            self._model = transformers.AutoModel.from_pretrained(self.model_id).to(self.device)
        >>>        return self._model
        >>>
        >>>    def predict(self, x):
        >>>        x = self.tokenizer(x, return_tensors="pt")
        >>>        return self.model(x)

        >>> # Creating rh.Module instance
        >>> model = Model(model_id="bert-base-uncased", device="cuda")
        >>> model = model.to(system="my_gpu")
        >>> model.predict("Hello world!")   # Runs on system in process
        >>> tok = model.remote.tokenizer    # Returns remote tokenizer
        >>> id = model.local.model_id       # Returns local model_id, if any
        >>> model_id = model.model_id       # Returns local model_id (not remote)
        >>> model.fetch()                   # Returns full remote module, including model and tokenizer
        >>>

    Example - creating a Module from an existing class, via the rh.module() factory method:
        >>> other_model = Model(model_id="bert-base-uncased", device="cuda").to("my_gpu", "my_process")
        >>>
        >>> # Another method: Create a module instance from an existing non-Module class using rh.module()
        >>> RemoteModel = rh.module(cls=BERTModel).to("my_gpu", "my_process")
        >>> remote_model = RemoteModel(model_id="bert-base-uncased", device="cuda").to(system="my_gpu")
        >>> remote_model.predict("Hello world!")  # Runs on system in process
        >>>
        >>> # You can also call remote class methods
        >>> other_model = RemoteModel.get_model_size("bert-base-uncased")

        >>> # Loading a module
        >>> my_local_module = rh.module(name="~/my_module")
        >>> my_s3_module = rh.module(name="@/my_module")
    """
    if name and not cls:
        # Try reloading existing module
        return Module.from_name(name, load_from_den=load_from_den, dryrun=dryrun)

    cls_pointers = Module._extract_pointers(cls)

    name = name or (
        cls_pointers[2] if cls_pointers else generate_default_name(prefix="module")
    )

    module_subclass = _module_subclass_factory(cls, cls_pointers)
    return module_subclass._module_init_only(
        dryrun=dryrun,
        pointers=cls_pointers,
        name=name,
    )


MODULE_METHODS = dir(Module)
